import pyranges.raymock as ray

import numpy as np
import pandas as pd
from ncls import NCLS


def _both_dfs(scdf, ocdf, how=False):

    assert how in "containment first".split() + [False, None]
    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    it = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    if not how:
        _self_indexes, _other_indexes = it.all_overlaps_both(
            starts, ends, indexes)
    elif how == "containment":
        _self_indexes, _other_indexes = it.all_containments_both(
            starts, ends, indexes)
    else:
        _self_indexes, _other_indexes = it.first_overlap_both(
            starts, ends, indexes)

    _self_indexes = _self_indexes
    _other_indexes = _other_indexes

    return scdf.reindex(_self_indexes), ocdf.reindex(_other_indexes)


@ray.remote
def _write_both(scdf, ocdf, kwargs):

    if scdf.empty or ocdf.empty:
        return None

    suffixes = kwargs["suffixes"]
    how = kwargs["how"]
    new_pos = kwargs["new_pos"]

    scdf, ocdf = _both_dfs(scdf, ocdf, how=how)
    nix = pd.Index(range(len(scdf)))
    scdf.index = nix
    ocdf.index = nix

    ocdf = ocdf.drop("Chromosome", axis=1)

    if not new_pos:
        df = scdf.join(ocdf, rsuffix=suffixes[1])

    elif new_pos == "intersection":

        new_starts = pd.Series(
            np.where(scdf.Start.values > ocdf.Start.values, scdf.Start,
                     ocdf.Start),
            index=scdf.index,
            dtype=np.long)

        new_ends = pd.Series(
            np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
            index=scdf.index,
            dtype=np.long)
        df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
        df.insert(1, "Start", new_starts)
        df.insert(2, "End", new_ends)
        df.rename(
            index=str,
            columns={
                "Chromosome" + suffixes[0]: "Chromosome",
                "Strand" + suffixes[0]: "Strand"
            },
            inplace=True)

    elif new_pos == "union":

        new_starts = pd.Series(
            np.where(scdf.Start.values < ocdf.Start.values, scdf.Start,
                     ocdf.Start),
            index=scdf.index,
            dtype=np.long)

        new_ends = pd.Series(
            np.where(scdf.End.values > ocdf.End.values, scdf.End, ocdf.End),
            index=scdf.index,
            dtype=np.long)
        df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
        df.insert(1, "Start", new_starts)
        df.insert(2, "End", new_ends)
        df.rename(
            index=str,
            columns={
                "Chromosome" + suffixes[0]: "Chromosome",
                "Strand" + suffixes[0]: "Strand"
            },
            inplace=True)

    return df



# pragma: no cover
import pandas as pd

import pyranges as pr


def pyrange_or_df(func):
    def extension(self, other, *args, **kwargs):
        df = func(self, other, *args, **kwargs)

        if kwargs.get("df"):
            return df

        return pr.PyRanges(df)

    return extension


def pyrange_or_df_single(func):

    # from pyranges import PyRanges
    def extension(self, *args, **kwargs):
        df = func(self, *args, **kwargs)

        if kwargs.get("df"):
            return df

        return pr.PyRanges(df)

    return extension


def _keep_transcript_with_most_exons(df):

    transcripts_with_most_exons = []

    for _, gdf in df.groupby("GeneID"):

        max_exon = gdf.ExonNumber.max()
        max_transcript = gdf.loc[gdf.ExonNumber == max_exon].Transcript.iloc[0]

        max_rows = gdf.loc[gdf.Transcript == max_transcript]

        transcripts_with_most_exons.append(max_rows)

    return pd.concat(transcripts_with_most_exons).reset_index(drop=True)


def tss_or_tes(df, which, slack=0):

    assert which in "tes tss".split()

    if "Feature" not in df:
        raise Exception("No Feature information in object.")

    _df = df[df.Feature == "transcript"]

    if which == "tss":
        _df = _tss(_df, slack)
    elif which == "tes":
        _df = _tes(_df, slack)

    return _df


def filter_transcripts(df, keep="most_exons"):

    return _keep_transcript_with_most_exons(df)


def _tss(df, slack=0, drop_duplicates=True):

    # try:
    #     df = self.df
    # except:
    #     df = self

    tss_pos = df.loc[df.Strand == "+"]

    tss_neg = df.loc[df.Strand == "-"].copy()

    # pd.options.mode.chained_assignment = None
    tss_neg.loc[:, "Start"] = tss_neg.End

    # pd.options.mode.chained_assignment = "warn"
    tss = pd.concat([tss_pos, tss_neg], sort=False)
    tss["End"] = tss.Start
    tss.End = tss.End + 1 + slack
    tss.Start = tss.Start - slack
    tss.loc[tss.Start < 0, "Start"] = 0

    if drop_duplicates:
        tss = tss.drop_duplicates("Chromosome Start End".split())

    tss.index = range(len(tss))

    return tss


def _tes(df, slack=0, drop_duplicates=True):

    # df = self.df

    tes_pos = df.loc[df.Strand == "+"]

    tes_neg = df.loc[df.Strand == "-"].copy()

    # pd.options.mode.chained_assignment = None
    tes_neg.loc[:, "Start"] = tes_neg.End

    # pd.options.mode.chained_assignment = "warn"
    tes = pd.concat([tes_pos, tes_neg], sort=False)
    tes["Start"] = tes.End
    tes.End = tes.End + 1 + slack
    tes.Start = tes.Start - slack
    tes.loc[tes.Start < 0, "Start"] = 0

    if drop_duplicates:
        tes = tes.drop_duplicates("Chromosome Start End".split())

    tes.index = range(len(tes))

    return tes


class GenomicFeaturesMethods():

    pr = None

    def __init__(self, pr):

        self.pr = pr

    @pyrange_or_df_single
    def tss(self, transcripts="all", drop_duplicate_tss=True, slack=0):

        pr = self.pr

        df = pr.df

        if transcripts == "all":
            pass
        elif transcripts == "most_exons":
            df = _keep_transcript_with_most_exons(df)

        if not pr.stranded:
            raise Exception(
                "Cannot compute TSSes or TESes without strand info. Perhaps use slack() instead?"
            )
        df = tss_or_tes(df, "tss", slack)

        if drop_duplicate_tss:
            df = df.drop_duplicates("Chromosome Start End".split())

        df = df.drop(["ExonNumber", "ExonID"], 1)

        return df

    @pyrange_or_df_single
    def tes(self, transcripts="all", drop_duplicate_tss=True, slack=0):

        pr = self.pr

        df = pr.df
        df = df.drop("ExonNumber", 1)

        if transcripts == "all":
            pass
        elif transcripts == "most_exons":
            df = _keep_transcript_with_most_exons(df)

        if not pr.stranded:
            raise Exception(
                "Cannot compute TSSes or TESes without strand info. Perhaps use slack() instead?"
            )
        df = tss_or_tes(df, "tes", slack)

        if drop_duplicate_tss:
            df = df.drop_duplicates("Chromosome Start End".split())

        df = df.drop(["ExonNumber", "ExonID"], 1)

        return df



from pyranges.subset import get_string, get_tuple, get_slice

from pyranges import PyRanges


def _getitem(self, val):
    if isinstance(val, str):
        df = get_string(self, val)
    elif isinstance(val, tuple):
        df = get_tuple(self, val)
    elif isinstance(val, slice):
        df = get_slice(self, val)
    else:
        raise Exception("Not valid subsetter: {}".format(str(val)))

    if not df is None:
        return PyRanges(df)
    else:
        return PyRanges({})



import os
import pyranges as pr

import tempfile

import requests


def get_ftp_file(ftp, binary_loc, path=None, test=False):

    url = "http://" + ftp + "/" + binary_loc
    if test:
        # if test only download small part
        from io import BytesIO
        import gzip
        headers = {"Range": "bytes=0-1000000"}
        c = requests.get(url, headers=headers).content
        s = gzip.GzipFile(fileobj=BytesIO(c)).read(1000000)
        f = s.decode().rsplit("\n", 1)[0]
        with tempfile.NamedTemporaryFile(suffix=".gtf", mode="w+") as t:
            t.write(f)
            gr = pr.read_gtf(t.name)
    else:
        if path and os.path.dirname(path):
            os.makedirs(os.path.dirname(path), exist_ok=True)

        c = requests.get(url).content

        if not path:
            with tempfile.NamedTemporaryFile(
                    suffix=".gtf.gz", mode="wb+") as t:
                t.write(c)
                gr = pr.read_gtf(t.name)
        else:
            if os.path.dirname(path):
                os.makedirs(os.path.dirname(path), exist_ok=True)
            fh = open(path, "wb+")
            fh.write(c)
            fh.close()

            gr = pr.read_gtf(path)

    return gr



from pyranges.multithreaded import pyrange_apply_single


def _coverage(ranges, value_col=None, strand=True, rpm=False):

    try:
        from pyrle.methods import coverage
        from pyrle import PyRles
    except ImportError:
        raise Exception(
            "Using the coverage method requires that pyrle is installed.")

    kwargs = {"value_col": value_col}
    if value_col is None:
        kwargs["sparse"] = {"self": True}
    # from pydbg import dbg

    result = pyrange_apply_single(coverage, ranges, strand, kwargs)

    if rpm:
        multiplier = 1e6 / len(ranges)
        result = {k: v * multiplier for k, v in result.items()}

    return PyRles(result)



from __future__ import print_function

import pyranges.raymock as ray

import sys

import pandas as pd
import numpy as np

import pkg_resources

from pyranges.pyranges import PyRanges
from pyranges.readers import read_gtf, read_bam, read_bed
from pyranges import data
from pyranges import db
from pyranges.methods.concat import concat

from pyrle import PyRles, Rle

from pyranges.version import __version__

get_example_path = data.get_example_path

read_gff = read_gtf



import numpy as np
import pandas as pd

from pyranges.statistics import StatisticsMethods
from pyranges.genomicfeatures import GenomicFeaturesMethods
from pyranges.out import OutMethods
from pyranges import PyRanges


def set_dtypes(df, extended):

    if not extended:
        dtypes = {
            "Start": np.int32,
            "End": np.int32,
            "Chromosome": "category",
            "Strand": "category"
        }
    else:
        dtypes = {
            "Start": np.int64,
            "End": np.int64,
            "Chromosome": "category",
            "Strand": "category"
        }

    if not "Strand" in df:
        del dtypes["Strand"]

    # on test data with a few rows, the below code does not work
    # therefore test for a good number of rows
    if len(df) > 500:
        categoricals = (df.nunique() / len(df) <= 0.1).replace({
            True: "category",
            False: "object"
        }).to_dict()
        categoricals.update(dtypes)
        dtypes = categoricals

    for col, dtype in dtypes.items():

        if df[col].dtype.name != dtype:
            df[col] = df[col].astype(dtype)

    return df


def create_df_dict(df):

    if "Strand" in df:
        grpby_key = "Chromosome Strand".split()
    else:
        grpby_key = "Chromosome"

    return {k: v for k, v in df.groupby(grpby_key)}


def create_pyranges_df(chromosomes, starts, ends, strands=None):

    if isinstance(chromosomes, str):
        chromosomes = pd.Series([chromosomes] * len(starts), dtype="category")

    if strands is not None:

        if isinstance(strands, str):
            strands = pd.Series([strands] * len(starts), dtype="category")

        columns = [chromosomes, starts, ends, strands]
        lengths = list(str(len(s)) for s in columns)
        assert len(
            set(lengths)
        ) == 1, "chromosomes, starts, ends and strands must be of equal length. But are {}".format(
            ", ".join(lengths))
        colnames = "Chromosome Start End Strand".split()
    else:
        columns = [chromosomes, starts, ends]
        lengths = list(str(len(s)) for s in columns)
        assert len(
            set(lengths)
        ) == 1, "chromosomes, starts and ends must be of equal length. But are {}".format(
            ", ".join(lengths))
        colnames = "Chromosome Start End".split()

    idx = range(len(starts))
    series_to_concat = []
    for s in columns:
        if isinstance(s, pd.Series):
            s = pd.Series(s.values, index=idx)
        else:
            s = pd.Series(s, index=idx)

        series_to_concat.append(s)

    df = pd.concat(series_to_concat, axis=1)
    df.columns = colnames

    return df


def _init(self,
          df=None,
          chromosomes=None,
          starts=None,
          ends=None,
          strands=None,
          copy_df=False,
          extended=False):

    if isinstance(df, PyRanges):
        raise Exception("Object is already a PyRange.")

    if copy_df:
        df = df.copy()

    if df is False or df is None:
        df = create_pyranges_df(chromosomes, starts, ends, strands)

    if isinstance(df, pd.DataFrame):
        df = set_dtypes(df, extended)

    # below is not a good idea! then gr["chr1"] might change the dtypes of a gr!
    # elif isinstance(df, dict):
    #     df = {k: set_dtypes(v, extended) for k, v in df.items()}

    if isinstance(df, pd.DataFrame):
        self.__dict__["dfs"] = create_df_dict(df)
    else:
        # df is actually dict of dfs
        self.__dict__["dfs"] = df

    self.__dict__["features"] = GenomicFeaturesMethods(self)
    self.__dict__["stats"] = StatisticsMethods(self)
    self.__dict__["out"] = OutMethods(self)



import shutil

import pandas as pd

from tabulate import tabulate


def get_terminal_size():

    return shutil.get_terminal_size().columns


def reduce_string_width(str_repr, s, h, n_intervals, n_chromosomes):

    terminal_width = get_terminal_size()

    hidden_cols = []
    header = str_repr.split("\n", 2)[1]

    while len(header) > terminal_width:
        header, hidden = header.rsplit("|", 1)
        hidden = hidden.strip()
        if hidden:
            hidden_cols.append(hidden.strip())

    columns = header.replace("|", "").split()

    str_repr = tabulate(
        s.get(columns), headers=h, tablefmt='psql', showindex=False)

    str_repr += "\nPyRanges object has {} sequences from {} chromosomes.".format(
        n_intervals, n_chromosomes)
    if hidden_cols:
        str_repr += "\nHidden columns: {}".format(", ".join(hidden_cols))

    return str_repr


def tostring(self):
    # print("in str")
    if len(self) == 0:
        return "Empty PyRanges"

    # keys = natsorted(list(self.dfs.keys()))
    if len(self.keys()) == 1:

        first_key = self.keys()[0]
        # last_key = list(self.dfs.keys())[-1]
        first_df = self.dfs[first_key]

        # last_df = ray.get(self.dfs[last_key]).tail(3)
        h = first_df.head(3).astype(object)
        m = first_df.head(1).astype(object)
        t = first_df.tail(3).astype(object)
        m.loc[:, :] = "..."

        if len(self) > 6:
            s = pd.concat([h, m, t])
        elif len(self) == 6:
            s = pd.concat([h, t])
        else:
            s = h
    else:
        keys = self.keys()
        first_key = keys[0]
        last_key = keys[-1]
        # first_key = self.keys[0]
        # last_key = self.keys[-1]
        first_df = self.dfs[first_key].head(3)
        last_df = self.dfs[last_key].tail(3)
        # last_df = self.dfs[list(self.dfs.keys())[-1]].tail(3)

        h = first_df.head(3).astype(object)
        m = first_df.head(1).astype(object)
        t = last_df.head(3).astype(object)
        m.loc[:, :] = "..."
        # m.index = ["..."]
        # print((len(h) + len(t)) < 6, len(self) >= 6)
        if (len(h) + len(t)) < 6:

            keys_covered = set()
            # iterate from front until have three
            heads = []
            hl = 0
            for k in keys:
                keys_covered.add(k)
                h = self.dfs[k].head(3)
                first_df = h
                hl += len(h)
                heads.append(h)
                if hl >= 3:
                    break

            tails = []
            tl = 0
            for k in keys[::-1]:
                if k in keys_covered:
                    continue

                t = self.dfs[k].tail(3)
                tl += len(t)
                tails.append(t)
                if tl >= 3:
                    break
            # iterate from back until have three

            h = pd.concat(heads).head(3).astype(object)
            if tails:
                t = pd.concat(tails).tail(3).astype(object)
                if len(h) + len(t) > 6:
                    m = h.head(1).astype(object)
                    m.loc[:, :] = "..."
                    s = pd.concat([h, m, t])
                else:
                    s = pd.concat([h, t])
            else:
                s = h

        elif len(h) + len(t) == 6:
            m.loc[:, :] = "..."
            s = pd.concat([h, m, t])
        else:
            s = pd.concat([h, t])

    if False:  # make setting
        if self.stranded:
            pos = s.Chromosome.astype(str) + " " + s.Start.astype(
                str) + "-" + s.End.astype(str) + " " + s.Strand.astype(str)
            s = s.drop("Chromosome Start End Strand".split(), axis=1)
            first_df = first_df.drop(
                "Chromosome Start End Strand".split(), axis=1)
        else:
            pos = s.Chromosome.astype(str) + " " + s.Start.astype(
                str) + "-" + s.End.astype(str)
            s = s.drop("Chromosome Start End".split(), axis=1)
            first_df = first_df.drop("Chromosome Start End".split(), axis=1)

        s.insert(0, "Position", pos)
        h = [
            c + "\n(" + str(t) + ")"
            for c, t in zip(s.columns, ["multiple types"] +
                            list(first_df.dtypes))
        ]
    else:
        dtypes = []
        for col, dtype in zip(s.columns, first_df.dtypes):
            if str(dtype) == "category":
                dtype = first_df[col].cat.codes.dtype

            # dtype = str(dtype).replace("float", "f_").replace("int", "i_")
            dtypes.append(dtype)

        h = [c + "\n(" + str(t) + ")" for c, t in zip(s.columns, list(dtypes))]

    str_repr = tabulate(s, headers=h, tablefmt='psql', showindex=False) + \
                                    "\nPyRanges object has {} sequences from {} chromosomes.".format(len(self), len(self.chromosomes))

    str_repr = reduce_string_width(str_repr, s, h, len(self),
                                   len(self.chromosomes))

    return str_repr



import pandas as pd
from natsort import natsorted

import pyranges as pr

import pyranges.raymock as ray

from pyranges.tostring import tostring

from pyranges.methods.intersection import _intersection, _overlap
from pyranges.multithreaded import pyrange_apply, pyrange_apply_single, _slack, _tes, _tss


def fill_kwargs(kwargs):

    defaults = {
        "strandedness": None,
        "overlap": True,
        "how": None,
        "invert": None,
        "new_pos": None,
        "suffixes": ["_a", "_b"],
        "suffix": "_b",
        "sparse": {
            "self": False,
            "other": False
        }
    }

    defaults.update(kwargs)

    return defaults


class PyRanges():

    dfs = None
    gf = None

    def __init__(self,
                 df=None,
                 chromosomes=None,
                 starts=None,
                 ends=None,
                 strands=None,
                 copy_df=False,
                 extended=False):

        from pyranges.methods.init import _init

        _init(self, df, chromosomes, starts, ends, strands, copy_df, extended)

    def __len__(self):
        return sum([len(d) for d in self.values()])

    def __call__(self, f, strand=False):
        return self.apply(f)

    def __getattr__(self, name):

        from pyranges.methods.attr import _getattr

        return _getattr(self, name)

    def __setattr__(self, column_name, column):

        from pyranges.methods.attr import _setattr

        _setattr(self, column_name, column)

    # def __eq__(self, other):

    #     return self.df.equals(other.df)

    def __getitem__(self, val):

        from pyranges.methods.getitem import _getitem

        return _getitem(self, val)

    def __str__(self):

        return tostring(self)

    def __repr__(self):

        return str(self)

    def overlap(self, other, **kwargs):

        kwargs["sparse"] = {"self": False, "other": True}
        kwargs = fill_kwargs(kwargs)

        dfs = pyrange_apply(_overlap, self, other, **kwargs)

        return PyRanges(dfs)

    # # TODO: use subtract code here instead, easier
    # def no_overlap(self, other, **kwargs):

    #     kwargs = fill_kwargs(kwargs)
    #     kwargs["invert"] = True
    #     kwargs["sparse"] = {"self": False, "other": True}

    #     # if kwargs["strandedness"] in ["same", "opposite"]:
    #     #     kwargs["strandedness"] = {
    #     #         "same": "opposite",
    #     #         "opposite": "same"
    #     #     }[kwargs["strandedness"]]
    #     dfs = pyrange_apply(_overlap, self, other, **kwargs)

    #     return PyRanges(dfs)

    def nearest(self, other, **kwargs):

        from pyranges.methods.nearest import _nearest

        kwargs = fill_kwargs(kwargs)

        dfs = pyrange_apply(_nearest, self, other, **kwargs)

        return PyRanges(dfs)

    def intersect(self, other, **kwargs):

        kwargs = fill_kwargs(kwargs)
        kwargs["sparse"] = {"self": False, "other": True}

        dfs = pyrange_apply(_intersection, self, other, **kwargs)

        return PyRanges(dfs)

    def set_intersect(self, other, **kwargs):

        kwargs = fill_kwargs(kwargs)
        strandedness = kwargs["strandedness"]
        strand = True if strandedness else False
        self_clusters = self.merge(strand=strand, **kwargs)
        other_clusters = other.merge(strand=strand, **kwargs)
        dfs = pyrange_apply(_intersection, self_clusters, other_clusters,
                            **kwargs)

        return PyRanges(dfs)

    def set_union(self, other, **kwargs):

        kwargs = fill_kwargs(kwargs)
        strandedness = kwargs["strandedness"]
        strand = True if strandedness else False

        gr = pr.concat([self, other])
        gr = gr.merge(strand=strand, **kwargs)

        return gr

    def subtract(self, other, **kwargs):

        from pyranges.methods.subtraction import _subtraction

        kwargs["sparse"] = {"self": False, "other": True}
        kwargs = fill_kwargs(kwargs)
        strandedness = kwargs["strandedness"]

        strand = True if strandedness else False
        other_clusters = other.merge(strand=strand, **kwargs)
        result = pyrange_apply(_subtraction, self, other_clusters, **kwargs)

        return PyRanges(result)

    def join(self, other, **kwargs):

        from pyranges.methods.join import _write_both

        kwargs = fill_kwargs(kwargs)
        dfs = pyrange_apply(_write_both, self, other, **kwargs)

        return PyRanges(dfs)

    def merge(self, strand=None, **kwargs):

        from pyranges.methods.merge import _merge

        kwargs["sparse"] = {"self": True}
        df = pyrange_apply_single(_merge, self, strand, kwargs)

        return PyRanges(df)

    def coverage(self, value_col=None, strand=False, rpm=False):

        from pyranges.methods.coverage import _coverage

        return _coverage(self, value_col, strand=strand, rpm=rpm)

    def apply(self, f, strand=False, as_pyranges=True, kwargs=None):

        if not kwargs:
            kwargs = {}
        kwargs = fill_kwargs(kwargs)

        f = ray.remote(f)

        result = pyrange_apply_single(f, self, strand, kwargs)

        if not as_pyranges:
            return result
        else:
            return PyRanges(result)

    def apply_pair(self, other, f, kwargs, strand=False, as_pyranges=True):

        f = ray.remote(f)

        result = pyrange_apply(f, self, other, **kwargs)

        if not as_pyranges:
            return result
        else:
            return PyRanges(result)

    def slack(self, slack):

        kwargs = fill_kwargs({"slack": slack})
        prg = PyRanges(
            pyrange_apply_single(_slack, self, self.stranded, kwargs))

        return prg

    def tssify(self, slack=0):

        kwargs = fill_kwargs({"slack": slack})
        return PyRanges(
            pyrange_apply_single(_tss, self, self.stranded, kwargs))

    def sort(self, columns=("Start", "End"), **kwargs):
        from pyranges.methods.sort import _sort
        kwargs["sparse"] = {"self": False}
        kwargs = fill_kwargs(kwargs)
        return PyRanges(
            pyrange_apply_single(_sort, self, self.stranded, kwargs))

    def tesify(self, slack=0):

        kwargs = fill_kwargs({"slack": slack})
        return PyRanges(
            pyrange_apply_single(_tes, self, self.stranded, kwargs))

    def keys(self):
        return natsorted(self.dfs.keys())

    @property
    def stranded(self):
        return len(list(self.keys())[0]) == 2

    @property
    def strands(self):

        if not self.stranded:
            raise Exception("PyRanges not stranded!")

        return natsorted(set([k[1] for k in self.keys()]))

    @property
    def chromosomes(self):

        if self.stranded:
            return natsorted(set([k[0] for k in self.keys()]))
        else:
            return natsorted(set([k for k in self.keys()]))

    def items(self):

        return natsorted([(k, df) for (k, df) in self.dfs.items()])

    def values(self):

        return [df for k, df in self.items() if not df.empty]

    @property
    def df(self):

        return self.as_df()

    def as_df(self):

        if len(self) == 0:
            return pd.DataFrame()
        elif len(self) == 1:
            return self.values()[0]
        else:
            return pd.concat(self.values())

    def lengths(self):

        lengths = {}
        for k, df in self.items():
            lengths[k] = df.End - df.Start

        return lengths

    # def midpoints(self):

    #     midpoints = {}
    #     for k, df in self.items():
    #         midpoints[k] = (df.End + df.Start) / 2

    #     return midpoints

    def summary(self):

        from pyranges.methods.summary import _summary

        return _summary(self)



from hypothesis.extra.pandas import data_frames, column, indexes
import hypothesis.strategies as st

from pyranges import PyRanges

import pandas as pd
import numpy as np

lengths = st.integers(min_value=1, max_value=int(1e7))
small_lengths = st.integers(min_value=1, max_value=int(1e4))

strands = st.sampled_from("+ -".split())
single_strand = st.sampled_from(["+"])
names = st.text("abcdefghijklmnopqrstuvxyz", min_size=1)
scores = st.integers(min_value=0, max_value=256)

datatype = st.sampled_from([pd.Series, np.array, list])

chromosomes = st.sampled_from(
    ["chr{}".format(str(e)) for e in list(range(1, 23)) + "X Y M".split()])
chromosomes_small = st.sampled_from(["chr1"])
cs = st.one_of(chromosomes, chromosomes_small)

runlengths = data_frames(
    index=indexes(dtype=np.int64, min_size=1, unique=True),
    columns=[
        column("Runs", st.integers(min_value=1, max_value=int(1e7))),
        # must have a min/max on floats because R S4vectors translates too big ones into inf.
        # which is unequal to eg -1.79769e+308 so the tests fail
        column("Values", st.integers(min_value=-int(1e7), max_value=int(1e7)))
    ])

better_dfs_min = data_frames(
    index=indexes(dtype=np.int64, min_size=1, unique=True, elements=lengths),
    columns=[
        column("Chromosome", cs),
        column("Start", elements=lengths),
        column("End", elements=small_lengths),
        # column("Name", elements=names),
        # column("Score", elements=scores),
        column("Strand", strands)
    ])

better_dfs_min_2 = data_frames(
    index=indexes(dtype=np.int64, min_size=2, unique=True, elements=lengths),
    columns=[
        column("Chromosome", chromosomes_small),
        column("Start", elements=lengths),
        column("End", elements=small_lengths),
        # column("Name", elements=names),
        # column("Score", elements=scores),
        column("Strand", single_strand)
    ])

better_dfs_min_single_chromosome = data_frames(
    index=indexes(dtype=np.int64, min_size=1, unique=True, elements=lengths),
    columns=[
        column("Chromosome", chromosomes_small),
        column("Start", elements=lengths),
        column("End", elements=small_lengths),
        # column("Name", elements=names),
        # column("Score", elements=scores),
        column("Strand", strands)
    ])

runlengths_same_length_integers = data_frames(
    index=indexes(dtype=np.int64, min_size=1, unique=True),
    columns=[
        column("Runs", st.integers(min_value=1, max_value=int(1e4))),
        column("Values", st.integers(min_value=1, max_value=int(1e4))),
        column("Values2", st.integers(min_value=1, max_value=int(1e4)))
    ])


@st.composite
def dfs_min2(draw):  # nosec
    df = draw(better_dfs_min_2)
    # strand = draw(use_strand)
    df.loc[:, "End"] += df.Start

    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)

    # if not strand:
    #     df = df.drop("Strand", axis=1)

    gr = PyRanges(df, extended=True)
    # gr = PyRanges(df)

    # do not sort like this, use pyranges sort
    # np.random.seed(draw(st.integers(min_value=0, max_value=int(1e6))))
    # gr.df = df.reindex(np.random.permutation(df.index.values))

    return gr


@st.composite
def dfs_min(draw):  # nosec
    df = draw(better_dfs_min)
    # strand = draw(use_strand)
    df.loc[:, "End"] += df.Start

    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)

    # stranded = draw(st.booleans())
    # if not strand:
    #     df = df.drop("Strand", axis=1)

    gr = PyRanges(df, extended=True)
    # gr = PyRanges(df)

    # do not sort like this, use pyranges sort
    # np.random.seed(draw(st.integers(min_value=0, max_value=int(1e6))))
    # gr.df = df.reindex(np.random.permutation(df.index.values))

    return gr


@st.composite
def df_data(draw):

    df = draw(better_dfs_min)
    df.loc[:, "End"] += df.Start

    chromosome_type = draw(datatype)
    start_type = draw(datatype)
    end_type = draw(datatype)
    strand_type = draw(datatype)

    strand = chromosome_type(df.Strand)
    chromosome = chromosome_type(df.Chromosome)
    start = chromosome_type(df.Start)
    end = chromosome_type(df.End)

    return chromosome, start, end, strand


@st.composite
def dfs_min_single_chromosome(draw):
    df = draw(better_dfs_min_single_chromosome)
    df.loc[:, "End"] += df.Start
    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)

    return df


@st.composite
def selector(draw):

    df = draw(better_dfs_min)
    h = df.head(1)
    chromosome = h["Chromosome"].iloc[0]
    start = h["Start"].iloc[0]
    end = h["End"].iloc[0]
    strand = h["Strand"].iloc[0]

    chromosome_bool = draw(st.booleans())
    strand_bool = draw(st.booleans())
    start_bool = draw(st.booleans())
    end_bool = draw(st.booleans())

    _slice = {
        (True, True): slice(start, end),
        (True, False): slice(start, None),
        (False, True): slice(None, end),
        (False, False): slice(None, None)
    }[start_bool, end_bool]

    to_return = []
    if chromosome_bool:
        to_return.append(chromosome)
    if strand_bool:
        to_return.append(strand)
    if start_bool or end_bool:
        to_return.append(_slice)

    return to_return



import pyranges as pr
import pandas as pd

from collections import defaultdict


def concat(pyranges, strand=False):

    grs_per_chromosome = defaultdict(list)

    if strand:
        assert all([
            gr.stranded for gr in pyranges
        ]), "Cannot do stranded concat, not all pyranges contain strand info."

    for gr in pyranges:
        for k, df in gr.dfs.items():
            grs_per_chromosome[k].append(df)

    new_pyrange = {}

    for k, v in grs_per_chromosome.items():
        new_pyrange[k] = pd.concat(v)

    return pr.PyRanges(new_pyrange)
